from typing import Tuple

from amino import do, List, _, Lists
from amino.do import Do
from amino.boolean import false, true
from amino.io import IOException

from ribosome.nvim.io import NS
from ribosome.trans.api import trans
from ribosome.trans.messages import Error, Info
from ribosome.logging import ribo_log
from ribosome.process import SubprocessResult
from ribosome.trans.effect import GatherSubprocs
from ribosome.trans.message_base import Message

from chromatin.components.core.logic import (install_plugins, add_installed, reboot_plugins, activate_by_names,
                                             deactivate_by_names)
from chromatin import Env
from chromatin.util import resources
from chromatin.venv import Venv

from amino import __


# TODO add trans effect for logging
@trans.free.one(trans.st)
@do(NS[Env, Message])
def install_result(installed: List[SubprocessResult[Venv]], errors: List[IOException]) -> Do:
    errors.foreach(lambda e: ribo_log.caught_exception('installing plugin', e))
    success, failed = installed.split(_.success)
    success_venvs = success / _.data
    failed_venvs = failed / _.data
    yield success_venvs.traverse(add_installed, NS)
    msg = failed.empty.c(
        lambda: Info(resources.installed_plugins(success_venvs / _.name)),
        lambda: Error(resources.plugins_install_failed(failed_venvs / _.name)),
    )
    yield NS.pure(msg)


@trans.free.result(trans.st, trans.gather_subprocs)
@do(NS[Env, GatherSubprocs[Venv, Tuple[List[Venv], List[IOException]]]])
def install_missing() -> Do:
    missing = yield NS.inspect_f(_.missing)
    yield install_plugins(missing, false)


@trans.free.multi(trans.st)
def activate(*plugins: str) -> NS[Env, None]:
    return activate_by_names(Lists.wrap(plugins))


@trans.free.multi(trans.st)
def deactivate(*plugins: str) -> NS[Env, None]:
    return deactivate_by_names(Lists.wrap(plugins))


@do(NS[Env, None])
def updated(result: List[Venv]) -> Do:
    venvs = result.filter_not(_.name == 'chromatin')
    autoreboot = yield NS.inspect_f(_.autoreboot)
    if autoreboot:
        yield reboot_plugins(venvs / _.name)


@trans.free.result(trans.st, trans.gather_subprocs)
@do(NS[Env, GatherSubprocs[Venv, Tuple[List[Venv], List[IOException]]]])
def update_plugins_io(plugins: List[str]) -> Do:
    venvs = yield NS.inspect(__.updateable(plugins))
    yield (
        NS.error(resources.no_plugins_match_for_update(plugins))
        if venvs.empty else
        install_plugins(venvs, true)
    )


@trans.free.one(trans.st)
@do(NS[Env, Message])
def updated_plugins(results: List[SubprocessResult[Venv]], errors: List[IOException]) -> Do:
    errors.foreach(lambda e: ribo_log.caught_exception('updating plugin', e))
    success, failed = results.split(_.success)
    success_venvs = success / _.data
    failed_venvs = failed / _.data
    yield updated(success_venvs)
    msg = failed.empty.c(
        lambda: Info(resources.updated_plugins(success_venvs / _.name)),
        lambda: Error(resources.plugins_install_failed(failed_venvs / _.name)),
    )
    yield NS.pure(msg)


@trans.free.do()
@do(NS[Env, None])
def update_plugins(*ps: str) -> Do:
    plugins = Lists.wrap(ps)
    success, fail = yield update_plugins_io(plugins).m
    yield updated_plugins(success, fail).m


@trans.free.unit()
def reboot(*plugins: str) -> NS[Env, None]:
    return reboot_plugins(Lists.wrap(plugins))

__all__ = ('install_result', 'install_missing', 'update_plugins')



from uuid import UUID, uuid4

from amino import List, Either, Map
from amino.dat import Dat


class RpluginSpec(Dat['RpluginSpec']):

    @staticmethod
    def cons(name: str, spec: str, id: UUID=None) -> 'RpluginSpec':
        return RpluginSpec(name, spec, id or uuid4())

    def __init__(self, name: str, spec: str, id: UUID) -> None:
        self.name = name
        self.spec = spec
        self.uuid = id

    @staticmethod
    def from_config(data: dict) -> Either[str, 'RpluginSpec']:
        m = Map(data)
        def create(spec: str) -> RpluginSpec:
            name = m.lift('name') | spec
            return RpluginSpec.cons(name=str(name), spec=str(spec))
        return m.lift('spec').to_either(f'plugin data {data} missing attribute `spec`').map(create)

    @staticmethod
    def simple(name: str) -> 'RpluginSpec':
        return RpluginSpec.cons(name=name, spec=name)

    def _arg_desc(self) -> List[str]:
        return List(self.name, self.spec)

__all__ = ('RpluginSpec',)



from typing import Tuple

from amino import do, __, _, Just, Maybe, List, Either, Nil, Boolean, Right, Left, Path, Lists
from amino.do import Do
from amino.state import State, EitherState
from amino.io import IOException
from amino.func import ReplaceVal
from amino.util.string import camelcaseify, camelcase

from ribosome.nvim.io import NS
from ribosome.trans.message_base import Message
from ribosome.trans.effect import GatherSubprocs
from ribosome.trans.send_message import transform_data_state
from ribosome.process import SubprocessResult
from ribosome.logging import ribo_log
from ribosome.request.rpc import DefinedHandler, RpcHandlerSpec
from ribosome.nvim import NvimIO

from chromatin import Env
from chromatin.plugin import RpluginSpec
from chromatin.venvs import VenvAbsent, VenvFacade
from chromatin.venv import Venv, ActiveVenv
from chromatin.host import start_host, stop_host
from chromatin.util import resources


@do(NvimIO[List[DefinedHandler]])
def define_handlers(active_venv: ActiveVenv) -> Do:
    venv = active_venv.venv
    name = venv.name
    channel = active_venv.channel
    cname = camelcaseify(name)
    rpc_handlers_fun = f'{cname}RpcHandlers'
    result = yield NvimIO.call_once_defined(rpc_handlers_fun, timeout=3)
    handlers = (
        Lists.wrap(result)
        .flat_map(RpcHandlerSpec.decode)
        .map(lambda spec: DefinedHandler(spec=spec, channel=channel))
    )
    yield NvimIO.pure(handlers)


@do(NS[Env, None])
def activated(active_venv: ActiveVenv) -> Do:
    ribo_log.debug(f'activated {active_venv}')
    venv = active_venv.venv
    yield NS.delay(__.runtime(f'chromatin/{venv.name}/*'))
    yield NS.modify(__.host_started(active_venv))
    handlers = yield NS.lift(define_handlers(active_venv))
    yield NS.modify(__.add_handlers(venv, handlers))


@do(NS[Env, ActiveVenv])
def start_venv_host(venv: Venv, python_exe: Path, bin_path: Path) -> Do:
    debug = yield NS.inspect_f(_.settings.debug_pythonpath.value_or_default)
    channel, pid = yield NS.lift(start_host(python_exe, bin_path, venv.plugin_path, debug))
    yield NS.pure(ActiveVenv(venv=venv, channel=channel, pid=pid))


@do(Either[str, NS[Env, ActiveVenv]])
def venv_activation_io(venv: Venv) -> Do:
    python_exe = yield venv.python_executable
    bin_path = yield venv.bin_path
    yield Right(start_venv_host(venv, python_exe, bin_path))


@do(NS[Env, None])
def activate_venv(venv: Venv) -> Do:
    active_env = yield venv_activation_io(venv).value_or(NS.failed)
    yield activated(active_env)


@do(NS[Env, None])
def activate_multi(venvs: List[Venv]) -> Do:
    active = yield NS.inspect(_.active_venvs)
    already_active, inactive = venvs.split(active.contains)
    yield inactive.traverse(activate_venv, NS)


@do(NS[Env, None])
def activate_by_names(plugins: List[str]) -> Do:
    getter = _.installed if plugins.empty else __.installed_by_name(plugins)
    venvs = yield NS.inspect(getter)
    yield (
        NS.error(resources.no_plugins_match_for_activation(plugins))
        if venvs.empty else
        activate_multi(venvs)
    )


def activate_all() -> NS[Env, None]:
    return activate_by_names(List())


@do(NS[Env, None])
def deactivate_venv(venv: ActiveVenv) -> Do:
    def undef(spec: RpcHandlerSpec) -> NvimIO[str]:
        return NvimIO.cmd(spec.undef_cmdline, verbose=True)
    @do(NvimIO[None])
    def run(handlers: List[RpcHandlerSpec]) -> Do:
        yield NvimIO.cmd(f'{camelcase(venv.name)}Quit')
        yield stop_host(venv.channel)
        yield handlers.traverse(undef, NvimIO)
        ribo_log.debug(f'deactivated {venv}')
    handlers = yield NS.inspect(__.handlers_for(venv.name))
    specs = (handlers | Nil) / _.spec
    yield NS.modify(__.deactivate_venv(venv))
    yield NS.lift(run(specs))


def deactivate_multi(venvs: List[ActiveVenv]) -> NS[Env, List[NvimIO[None]]]:
    return venvs.traverse(deactivate_venv, NS)


@do(NS[Env, List[NvimIO[None]]])
def deactivate_by_names(plugins: List[str]) -> Do:
    getter = _.active if plugins.empty else __.active_by_name(plugins)
    venvs = yield NS.inspect(getter)
    yield (
        NS.error(resources.no_plugins_match_for_deactivation(plugins))
        if venvs.empty else
        deactivate_multi(venvs)
    )


@do(NS[Env, None])
def reboot_plugins(plugins: List[str]) -> Do:
    yield activate_by_names(plugins)
    yield deactivate_by_names(plugins)


@do(NS[Env, None])
def activation_complete() -> Do:
    venvs = yield NS.inspect(_.uninitialized)
    prefixes = venvs / _.name / camelcase
    def stage(num: int) -> NvimIO[None]:
        return prefixes.traverse(lambda a: NvimIO.cmd_sync(f'{a}Stage{num}'), NvimIO)
    yield NS.lift(Lists.range(1, 5).traverse(stage, NvimIO))
    yield NS.modify(__.initialization_complete())


@do(NS[Env, None])
def activate_newly_installed() -> Do:
    new = yield NS.inspect(_.inactive)
    yield activate_multi(new)
    yield activation_complete()


def add_venv(venv: Venv) -> State[Env, None]:
    return State.modify(__.add_venv(venv))


def add_installed(venv: Venv) -> State[Env, None]:
    return State.modify(__.add_installed(venv))


def bootstrap(venv_facade: VenvFacade, venv: VenvAbsent) -> None:
    return venv_facade.bootstrap(venv.plugin)


@do(EitherState[Env, None])
def venv_setup_result(result: List[Either[str, Venv]]) -> Do:
    def split(z: Tuple[List[str], List[Venv]], a: Either[str, Venv]) -> Tuple[List[str], List[Venv]]:
        err, vs = z
        return a.cata((lambda e: (err.cat(e), vs)), (lambda v: (err, vs.cat(v))))
    errors, venvs = result.fold_left((Nil, Nil))(split)
    yield transform_data_state(venvs.traverse(add_venv, State)).to(EitherState)
    error = errors.map(str).join_comma
    ret = Left(f'failed to setup venvs: {error}') if errors else Right(None)
    yield EitherState.lift(ret)


def install_plugins_result(results: List[Either[IOException, SubprocessResult[Venv]]]) -> Maybe[Message]:
    venvs = results.flat_map(__.cata(ReplaceVal(Nil), List))
    errors = results.flat_map(__.cata(List, ReplaceVal(Nil)))
    return venvs, errors


# FIXME why not parallel?
@do(NS[Env, GatherSubprocs[Venv, Tuple[List[Venv], List[IOException]]]])
def install_plugins(venvs: List[Venv], update: Boolean) -> Do:
    '''run subprocesses in sequence that install packages into their venvs using pip.
    cannot be run in parallel as they seem to deadlock.
    '''
    venv_facade = yield NS.inspect_f(_.venv_facade)
    @do(Either[str, GatherSubprocs])
    def plugin_venvs(env: Env) -> Do:
        pvenvs = yield venvs.traverse(env.plugin_venv, Either)
        procs = yield pvenvs.traverse(venv_facade.install, Either)
        yield Right(GatherSubprocs(procs, install_plugins_result, timeout=60))
    pvenvs = yield NS.inspect(plugin_venvs)
    yield NS.from_either(pvenvs)


@do(NS[Env, None])
def add_crm_venv() -> Do:
    handle = yield NS.inspect_f(_.handle_crm)
    if handle:
        plugin = RpluginSpec.simple('chromatin')
        yield NS.modify(__.set.chromatin_plugin(Just(plugin)))
        venv_facade = yield NS.inspect_f(_.venv_facade)
        venv = venv_facade.cons(plugin)
        yield NS.modify(__.set.chromatin_venv(Just(venv)))


@do(NS[Env, List[RpluginSpec]])
def read_conf() -> Do:
    plugin_conf = yield NS.inspect_f(_.rplugins)
    specs = plugin_conf.traverse(RpluginSpec.from_config, Either)
    yield NS.from_either(specs)


__all__ = ('add_crm_venv', 'read_conf', 'install_plugins', 'bootstrap', 'add_installed', 'defined_handlers')



from types import SimpleNamespace
from typing import Any

from amino import Path, List, Try, Either
from amino.dat import Dat

from chromatin.plugin import RpluginSpec


class Venv(Dat['Venv']):

    def __init__(
            self,
            name: str,
            dir: Path,
            python_executable: Either[str, Path],
            bin_path: Either[str, Path]
    ) -> None:
        self.name = name
        self.dir = dir
        self.python_executable = python_executable
        self.bin_path = bin_path

    @staticmethod
    def from_ns(dir: Path, plugin: RpluginSpec, context: SimpleNamespace) -> 'Venv':
        exe = Try(lambda: context.env_exe) / Path
        bin_path = Try(lambda: context.bin_path) / Path
        return Venv(name=plugin.name, dir=dir, python_executable=exe, bin_path=bin_path)

    def _arg_desc(self) -> List[str]:
        return List(str(self.name), str(self.dir))

    @property
    def site(self) -> Path:
        return self.dir / 'lib' / 'python3.6' / 'site-packages'

    @property
    def plugin_path(self) -> Path:
        return self.site / self.name / '__init__.py'


class PluginVenv(Dat['PluginVenv']):

    def __init__(self, venv: Venv, plugin: RpluginSpec) -> None:
        self.venv = venv
        self.plugin = plugin

    @property
    def req(self) -> str:
        return self.plugin.spec

    @property
    def name(self) -> str:
        return self.plugin.name

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.venv, self.plugin)


class ActiveVenv(Dat['ActiveVenv']):

    def __init__(self, venv: Venv, channel: int, pid: int) -> None:
        self.venv = venv
        self.channel = channel
        self.pid = pid

    @property
    def name(self) -> str:
        return self.venv.name

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.venv, self.channel, self.pid)

__all__ = ('Venv', 'ActiveVenv', 'PluginVenv')



import typing
from typing import Tuple

from ribosome.nvim import NvimIO
from ribosome.logging import ribo_log

from amino import Path, do, __, List
from amino.do import Do


stderr_handler_name = 'ChromatinJobStderr'

stderr_handler_body = '''
  echoerr 'error in chromatin rpc job on channel ' . a:id . ': ' . string(a:data)
'''


def host_cmdline(python_exe: Path, bin_path: Path, plug: Path, debug: bool) -> typing.List[str]:
    debug_option = [] if debug else ['-E']
    args = [str(bin_path / f'ribosome_start_plugin'), str(plug)]
    return [str(python_exe)] + debug_option + args


@do(NvimIO[None])
def define_stderr_handler() -> Do:
    exists = yield NvimIO.delay(__.function_exists(stderr_handler_name))
    if not exists:
        yield NvimIO.delay(__.define_function(stderr_handler_name, List('id', 'data', 'event'), stderr_handler_body))


@do(NvimIO[Tuple[int, int]])
def start_host(python_exe: Path, bin_path: Path, plugin_path: Path, debug: bool=False) -> Do:
    yield define_stderr_handler()
    cmdline = host_cmdline(python_exe, bin_path, plugin_path, debug)
    ribo_log.debug(f'starting host: {cmdline}')
    channel = yield NvimIO.call('jobstart', cmdline, dict(rpc=True, on_stderr=stderr_handler_name))
    pid = yield NvimIO.call('jobpid', channel)
    ribo_log.debug(f'host running, channel {channel}, pid {pid}')
    yield NvimIO.pure((channel, pid))


def stop_host(channel: int) -> NvimIO[None]:
    return NvimIO.call('jobstop', channel)

__all__ = ('start_host', 'stop_host')



from amino import do, __, List, _, Boolean, L, Map, Path
from amino.do import Do
from amino.state import State

from ribosome.nvim.io import NS
from ribosome.trans.api import trans
from ribosome.trans.message_base import Message
from ribosome.trans.action import TransM
from ribosome.trans.effect import GatherIOs
from ribosome.trans.messages import Info

from chromatin.components.core.logic import (add_crm_venv, read_conf, activate_newly_installed, bootstrap,
                                             venv_setup_result, add_installed)
from chromatin import Env
from chromatin.plugin import RpluginSpec
from chromatin.venvs import VenvExistent
from chromatin.components.core.trans.install import install_missing, install_result


@trans.free.result(trans.st)
@do(NS[Env, Message])
def init() -> Do:
    yield add_crm_venv()
    yield NS.delay(__.vars.set_p('started', True))
    yield NS.delay(__.vars.ensure_p('rplugins', []))
    yield read_conf()


@trans.free.unit(trans.st, trans.gather_ios, trans.st)
@do(NS[Env, GatherIOs])
def setup_venvs() -> Do:
    '''check whether a venv exists for each plugin in the env.
    for those that don't, create venvs in `g:chromatin_venv_dir`.
    '''
    venv_facade = yield NS.inspect_f(_.venv_facade)
    plugins = yield NS.inspect(_.plugins)
    jobs = plugins / venv_facade.check
    existent, absent = jobs.split_type(VenvExistent)
    yield existent.traverse(add_installed, State).nvim
    yield NS.pure(GatherIOs(absent.map(L(bootstrap)(venv_facade, _)), venv_setup_result, timeout=30))


@trans.free.unit(trans.st)
def post_setup() -> NS[Env, List[Message]]:
    return activate_newly_installed()


@trans.free.do()
@do(TransM)
def setup_plugins() -> Do:
    yield setup_venvs.m
    installed, errors = yield install_missing.m
    yield install_result(installed, errors).m
    yield post_setup.m


@trans.free.result(trans.st)
@do(NS[Env, Boolean])
def add_plugins(plugins: List[RpluginSpec]) -> None:
    yield NS.modify(__.add_plugins(plugins))
    yield NS.inspect_f(_.autostart)


@trans.free.do()
@do(TransM)
def stage_1() -> Do:
    plugins = yield init.m
    autostart = yield add_plugins(plugins).m
    if autostart:
        yield setup_plugins.m


def show_plugins_message(venv_dir: Path, plugins: List[RpluginSpec]) -> str:
    venv_dir_msg = f'virtualenv dir: {venv_dir}'
    plugins_desc = plugins.map(_.spec).cons('Configured plugins:')
    return plugins_desc.cons(venv_dir_msg).join_lines


@trans.free.one(trans.st)
@do(NS[Env, Message])
def show_plugins() -> Do:
    venv_dir = yield NS.inspect_f(_.venv_dir)
    yield NS.inspect(lambda data: Info(show_plugins_message(venv_dir, data.plugins)))


@trans.free.do()
@do(NS[Env, None])
def add_plugin(spec: str, name=None) -> Do:
    name = name or spec
    plugin = RpluginSpec.cons(name, spec)
    autostart = yield add_plugins(List(plugin)).m
    if autostart:
        yield setup_plugins.m


__all__ = ('stage_1', 'show_plugins')



import sys
import inspect
import traceback

from ribosome.plugin import AutoPlugin, plugin_class_from_config


def echo(nvim: 'neovim.api.Nvim', msg: str) -> None:
    safe = msg.replace('"', '\\"')
    nvim.command(f'echo "{safe}"')


def run() -> int:
    try:
        import neovim
        nvim = neovim.attach('stdio')
        try:
            from amino.logging import amino_root_logger, amino_root_file_logging, TEST
            amino_root_file_logging()
            amino_root_logger.debug('starting chromatin, stage amino')
            try:
                from ribosome.nvim import NvimFacade
                from ribosome.logging import nvim_logging
                nvim_facade = NvimFacade(nvim, 'define_handlers').proxy
                nvim_logging(nvim_facade, TEST)
                amino_root_logger.debug('starting chromatin, stage ribosome')
                return stage2(nvim_facade)
            except Exception as e:
                amino_root_logger.caught_exception_error(f'importing ribosome in chromatin bootstrap', e)
                raise
        except Exception as e:
            msg = traceback.format_exc()
            echo(nvim, msg)
            raise
    except Exception:
        msg = traceback.format_exc()
        try:
            import amino
        except Exception as e:
            pass
        else:
            if amino.development:
                print(msg, file=sys.stderr)
        return 1


def stage2(nvim: 'ribosome.NvimFacade') -> int:
    try:
        from amino import Path, Lists
        from ribosome.logging import ribo_log
        from ribosome.rpc import rpc_handlers, define_handlers
        import chromatin
        from chromatin.host import start_host
        ex, bp, ins = Lists.wrap(sys.argv).lift_all(1, 2, 3).get_or_fail(f'invalid arg count for `crm_run`: {sys.argv}')
        python_exe = Path(ex)
        bin_path = Path(bp)
        installed = ins == '1'
        ribo_log.debug(f'starting chromatin, stage 2: {sys.argv}')
        plugin = plugin_class_from_config(chromatin.config, AutoPlugin, False)
        plugin_path = chromatin.__file__
        channel, pid = start_host(python_exe, bin_path, plugin_path).attempt(nvim).get_or_raise
        ribo_log.debug(f'starting chromatin, host started: {channel}/{pid}')
        handlers = rpc_handlers(plugin)
        define_handlers(channel, handlers, 'chromatin', str(plugin_path)).attempt(nvim).get_or_raise
        ribo_log.debug('starting chromatin, handlers defined')
        if installed:
            ribo_log.info('chromatin initialized. installing plugins...')
        def error(a: str) -> None:
            raise Exception(f'failed to initialize chromatin: {a}')
        nvim.cmd_once_defined('ChromatinStage1').leffect(error)
        return 0
    except Exception as e:
        ribo_log.caught_exception_error('initializing chromatin', e)
        raise


__all__ = ('run',)



from amino import List, Map

from ribosome.config import Config
from ribosome.request.handler.handler import RequestHandler
from ribosome.request.handler.prefix import Plain, Full

from chromatin.env import Env
from chromatin.settings import CrmSettings
from chromatin.components.core.main import Core
from chromatin.components.core.trans.install import update_plugins, activate, deactivate, reboot
from chromatin.components.core.trans.setup import stage_1, setup_plugins, show_plugins, add_plugin

config: Config = Config.cons(
    name='chromatin',
    prefix='crm',
    state_ctor=Env.cons,
    components=Map(core=Core),
    settings=CrmSettings(),
    # core_components=List('core'),
    request_handlers=List(
        RequestHandler.trans_cmd(stage_1)(prefix=Full()),
        RequestHandler.trans_cmd(add_plugin)(name='cram', prefix=Plain()),
        RequestHandler.trans_cmd(setup_plugins)(),
        RequestHandler.trans_cmd(show_plugins)(),
        RequestHandler.trans_cmd(activate)(),
        RequestHandler.trans_cmd(deactivate)(),
        RequestHandler.trans_cmd(reboot)(),
        RequestHandler.trans_cmd(update_plugins)(name='update'),
    )
)

__all__ = ('config',)



from typing import Any

from ribosome.request.rpc import DefinedHandler
from ribosome.config import Data, Config, PluginSettings
from ribosome.nvim import NvimIO

from amino import List, _, Either, Path, Right, do, Boolean, Map, __, Maybe, Nil, Nothing
from amino.boolean import true, false
from amino.dat import Dat
from amino.do import Do

from chromatin.logging import Logging
from chromatin.plugin import RpluginSpec
from chromatin.venv import Venv, ActiveVenv, PluginVenv
from chromatin.venvs import VenvFacade
from chromatin.settings import CrmSettings


def filter_venvs_by_name(venvs: List[Venv], names: List[str]) -> List[Venv]:
    return venvs if names.empty else venvs.filter(lambda v: v.name in names)


class Env(Dat['Env'], Logging, Data):

    @staticmethod
    def cons(config: Config[PluginSettings, 'Env']) -> 'Env':
        return Env(config, false, Nil, Nothing, Nothing, Map(), Nil, Nil, Nil, Map())

    def __init__(
            self,
            config: Config,
            initialized: Boolean,
            plugins: List[RpluginSpec],
            chromatin_plugin: Maybe[RpluginSpec],
            chromatin_venv: Maybe[Venv],
            venvs: Map[str, Venv],
            installed: List[Venv],
            active: List[ActiveVenv],
            uninitialized: List[ActiveVenv],
            handlers: Map[str, List[DefinedHandler]],
    ) -> None:
        self.config = config
        self.initialized = initialized
        self.plugins = plugins
        self.chromatin_plugin = chromatin_plugin
        self.chromatin_venv = chromatin_venv
        self.venvs = venvs
        self.installed = installed
        self.active = active
        self.uninitialized = uninitialized
        self.handlers = handlers

    @property
    def settings(self) -> CrmSettings:
        return self.config.settings

    def add_plugin(self, name: str, spec: str) -> 'Env':
        return self.append1.plugins(RpluginSpec.cons(name=name, spec=spec))

    def add_plugins(self, plugins: List[RpluginSpec]) -> 'Env':
        return self.append.plugins(plugins)

    def add_venv(self, venv: Venv) -> 'Env':
        return self.append.venvs((venv.name, venv))

    def add_installed(self, venv: Venv) -> 'Env':
        return self.append1.installed(venv)

    @property
    def plugins_with_crm(self) -> List[RpluginSpec]:
        return self.plugins.cons_m(self.chromatin_plugin)

    def plugin_by_name(self, name: str) -> Either[str, RpluginSpec]:
        return self.plugins_with_crm.find(lambda a: a.name == name).to_either(f'no plugin with name `{name}`')

    @do(Either[str, PluginVenv])
    def plugin_venv(self, venv: Venv) -> Do:
        plugin = yield self.plugin_by_name(venv.name)
        yield Right(PluginVenv(venv=venv, plugin=plugin))

    def missing_in(self, venvs: VenvFacade) -> List[Venv]:
        return self.venvs.v.filter_not(venvs.package_installed)

    @property
    def missing(self) -> NvimIO[List[Venv]]:
        return self.venv_facade / self.missing_in

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.plugins, self.venvs)

    @property
    def rplugins(self) -> NvimIO[List[Map[str, str]]]:
        return self.settings.rplugins.value_or_default

    @property
    def venv_dir(self) -> NvimIO[Path]:
        return self.settings.venv_dir.value_or_default

    @property
    def venv_facade(self) -> NvimIO[VenvFacade]:
        return self.venv_dir / VenvFacade

    @property
    def autostart(self) -> NvimIO[Boolean]:
        return self.settings.autostart.value_or_default

    @property
    def handle_crm(self) -> NvimIO[Boolean]:
        return self.settings.handle_crm.value_or_default

    @property
    def autoreboot(self) -> NvimIO[Boolean]:
        return self.settings.autoreboot.value_or_default

    @property
    def active_venvs(self) -> List[Venv]:
        return self.active / _.venv

    @property
    def inactive(self) -> List[ActiveVenv]:
        return self.installed.remove_all(self.active_venvs)

    def host_started(self, venv: ActiveVenv) -> 'Env':
        return self.append1.active(venv).append1.uninitialized(venv)

    def initialization_complete(self) -> 'Env':
        return self.set.uninitialized(List())

    def deactivate_venv(self, venv: ActiveVenv) -> 'Env':
        return self.mod.active(__.without(venv))

    def installed_by_name(self, names: List[str]) -> List[Venv]:
        return self.installed.filter(lambda v: v.name in names)

    def active_by_name(self, names: List[str]) -> List[str]:
        return self.active.filter(lambda v: v.name in names)

    def add_handlers(self, venv: Venv, handlers: List[DefinedHandler]) -> 'Env':
        return self.append.handlers((venv.name, handlers))

    def handlers_for(self, plugin: str) -> Either[str, List[DefinedHandler]]:
        return self.handlers.lift(plugin).to_either(f'no handlers defined for {plugin}')

    @property
    def installed_with_crm(self) -> List[Venv]:
        return self.installed.cons_m(self.chromatin_venv)

    def updateable(self, plugins: List[str]) -> List[Venv]:
        return filter_venvs_by_name(self.installed_with_crm, plugins)


__all__ = ('Env',)



from amino import Path, Try, Right, Either, do, Nil
from amino.do import Do
from amino.boolean import true, false

from ribosome.config import PluginSettings
from ribosome.config.settings import bool_setting, path_setting, list_setting

from chromatin.util.resources import xdg_cache_home, create_venv_dir_error

handle_crm_help = '''When updating plugins, chromatin can also update itself. To take effect, neovim has to be
restarted.
'''
venv_dir_help = '''Chromatin stores installed plugins in separate virtualenvs located in subdirectories of this setting.
'''
rplugin_help = '''The remote plugins managed by chromatin can be configured as a list of dictionaries of the form:
{
    'name': 'title',
    'spec': 'title==0.1.1',
}
`spec` can be any specifier that `pip` understands. If omitted, `name` is used.
'''
autostart_help = '''When set, chromatin will commence installation and activation of plugins on start.
'''
debug_pythonpath_help = '''When set, rplugins will bei started with the outer python env, effectively keeping the
$PYTHONPATH env var as it is in neovim.
'''
autoreboot_help = '''When set, plugins will be de- and reactivated after being updated.
'''


@do(Either[str, Path])
def default_venv_dir() -> Do:
    xdg_cache_path = xdg_cache_home.value / Path | (Path.home() / '.cache')
    venv_dir = xdg_cache_path / 'chromatin' / 'venvs'
    yield Try(venv_dir.mkdir, parents=True, exist_ok=True).lmap(lambda a: create_venv_dir_error(venv_dir))
    yield Right(venv_dir)


class CrmSettings(PluginSettings):

    def __init__(self) -> None:
        super().__init__('chromatin')
        self.handle_crm = bool_setting('handle_crm', 'update chromatin', handle_crm_help, True, Right(true))
        self.venv_dir = path_setting('venv_dir', 'virtualenv base directory', venv_dir_help, True, default_venv_dir())
        self.rplugins = list_setting('rplugins', 'rplugin config', rplugin_help, True, Right(Nil))
        self.autostart = bool_setting('autostart', 'autostart plugins', autostart_help, True, Right(true))
        self.debug_pythonpath = bool_setting('debug_pythonpath', 'pass through $PYTHONPATH', debug_pythonpath_help,
                                             True, Right(false))
        self.autoreboot = bool_setting('autoreboot', 'autoreboot plugins', autoreboot_help, True, Right(true))


__all__ = ('CrmSettings',)



from lenses import Lens, lens

from ribosome.dispatch.component import Component

from amino.state import State
from amino import Nothing

from chromatin.env import Env


class Core(Component):
    pass

    def state_lens(self, tpe: str, name: str) -> State[Env, Lens]:
        return (
            State.inspect(lambda s: s.plugins.index_where(lambda a: a.name == name) / (lambda i: lens(s).plugins[i]))
            if tpe == 'vim_plugin' else
            State.pure(Nothing)
        )


__all__ = ('Core',)
